/* martian.c: pci kernel mode driver module
 * 	
 * description:
 *	driver takes control over 164x Agere devices
 *	real work only in pair with martian_modem
 *
 * author(s):
 *	A. Chentsov <chentsov at ukr.net> 
 *
 * copying: GPLv2
 *	
 *
 * provides some inevitable low-level support for the user space modem driver (martian_modem)
 *
 * communication actions: 
 *	- pass device configuration information to the martian_modem
 *	- process interrupts, notify driver through poll/read
 *
 * implementation details:
 *	every device is represented by the node in the vfs
 *	currently proc entries are used
 *	module and martian_modem have the common block of information
 *	the data is mapped into martian_modem process by mmap call
 *	this is forced by isr dependance on run-time core state
 *
 * valuable information sources:
 *	User mode drivers: interrupt handling
 *	Using interrupts 11 to 27, Larry Doolittle
 *	drivers/char/mem.c, driver/char/mmtimer.c
 *
 * changes:
 *	2006/11/28	ioctl for debug setting on user request
 *	2006/11/20	ioctls for simulating module failures
 *	2006/11/20	relaying fault to user, detectlet for common data overwrite
 *	2006/10/20	intermode synchronization
 *	2006/10/02	martian devices create/delete separated from pci driver interface
 *	2006/10/02	strenthened conditions for xircom devices
 *	2006/10/02	vma flags correction for different kernel versions
 *	2006/07/23	Rewritten fifos handling
 *	2006/07/23	Same for pdm rx
 *	2006/07/17 	Added full irq servicing for tx/rx
*/

#include <linux/module.h>
#include <linux/version.h>
#include <linux/gfp.h>		/*** __get_free_page, remap_pfn_range */

#include <linux/bitops.h>	/*** test_and_set_bit */

#include <linux/poll.h>		/*** POLLIN .. */
#include <linux/interrupt.h>

#ifdef CONFIG_COMPAT
#  include <linux/ioctl32.h>
#endif

#if defined (CONFIG_SMP) && (defined (__i386__) || defined (__x86_64__))
#  include <linux/delay.h>
#endif

#include "kernel/mixspinlock.h"
#include "kmartian.h"
#include "martian.h"
#include "marsio.h"

/***** Driver globals *****/

struct proc_dir_entry *martians_proc_dir = NULL;

/**************************/

// statistics
unsigned int interrupt_counter = 0;
static char irqstring[11] = "|\0\0\0\0\0\0\0\0\0\0";
static int irqptr = 0;

#define MARTIAN_IRQ_ACCOUNT(letter, count_letter) \
	irqstring[irqptr] = letter; mdev->icounts.count_letter++;

#define MARTIAN_IRQ_ACCOUNT_FIN \
	do { \
		if (++irqptr == 10) irqptr = 0; \
		irqstring[irqptr] = '|'; \
	} while (0)


/***** ISR *****/

/* martian_isr
 *	 do the basic io detecting/disabling
 */
static
irqreturn_t martian_isr (int irq, void *dev)
{
	struct martian *mdev = (struct martian *) dev;
	struct martian_common *mcb = mdev->common;
	u8 rd7;
	irqreturn_t retval = IRQ_NONE;
#if defined (CONFIG_SMP) && (defined (__i386__) || defined (__x86_64__))
	int use_lock = 0;
#endif

	interrupt_counter++;

	if (test_bit (KMSTATE_UREADY, &mdev->state) && mcb->dp_bamil_rd7 == 0xff) {
		MARTIAN_IRQ_ACCOUNT('b', b); 
		MARTIAN_IRQ_ACCOUNT_FIN; 
		return IRQ_NONE;
	}

#if defined (CONFIG_SMP) && (defined (__i386__) || defined (__x86_64__))
	if (test_bit (KMSETTING_TRUESMP, & mdev->state)) {
		mspin_lock (&mcb->lock);
		use_lock = 1;
	}
#endif

	/* read rd7 */
	rd7 = mars_read_register (mcb, 0xd7);

	if (rd7 == 0) {
		outb (mcb->BaseValue, mcb->BaseAddressIndex);
		MARTIAN_IRQ_ACCOUNT('n', n); 
	}
	else if (rd7 == 0xff) {
		outb (mcb->BaseValue, mcb->BaseAddressIndex);
		MARTIAN_IRQ_ACCOUNT('f', f); 
	}
	else if (rd7 & IR_SIGNIFICANT) {
		if (! test_bit (KMSTATE_UREADY, &mdev->state)) {
			mars_write_register (mcb, IR, 0xff);
			retval = IRQ_HANDLED;

			MARTIAN_IRQ_ACCOUNT('i', i); 
			goto fin;
		}

		if (mcb->x_dsp_mars == 1)
			mars_write_word (mcb, 0x3c, 0x101);

		/* disabling */
		mars_write_register (mcb, IR, 0xff);

		/* check not fully served yet by user */
		if (atomic_read (&mcb->UnservedIrqs) > 0) {
			mcb->OverlappedIrqs++;
			atomic_set (&mcb->UnservedIrqs, 0);
		}

		atomic_inc (&mcb->UnservedIrqs);

		/* check not passed yet to user */
		if (atomic_read (&mcb->UnpassedIrqs) > 0)
			mcb->RepeatedIrqs++;
		atomic_inc (&mcb->UnpassedIrqs);


		if (test_bit (KMSETTING_SERVE_IRQS, &mdev->state)) {
			if (rd7 & 0x4) 
				process_stream (mdev) && (rd7 ^= 4);
		}

		if ((rd7 & IR_SIGNIFICANT) || test_bit (KMSTATE_FIFO_FAULT, &mdev->state)) {
			// if (rd7 & IR_SIGNIFICANT)
			mcb->rd7_isr = rd7;
			/* notify readers */
			atomic_inc (& mdev->count);
			wake_up_interruptible (& mdev->wq);
		}
		else {
			mars_write_register (mcb, IR, mcb->dp_bamil_rd7);
			outb (mcb->BaseValue, mcb->BaseAddressIndex);

			atomic_dec (&mcb->UnservedIrqs);
			atomic_dec (&mcb->UnpassedIrqs);
		}

		retval = IRQ_HANDLED;

		MARTIAN_IRQ_ACCOUNT('i', i); 

	}
	else {
		mars_write_register (mcb, IR, mcb->dp_bamil_rd7);
		outb (mcb->BaseValue, mcb->BaseAddressIndex);
		
		retval = IRQ_HANDLED;
		MARTIAN_IRQ_ACCOUNT('o', o);
	}

fin:
#if defined (CONFIG_SMP) && (defined (__i386__) || defined (__x86_64__))
	if (use_lock)
		mspin_unlock (&mcb->lock);
#endif

	MARTIAN_IRQ_ACCOUNT_FIN; 

	return retval;
}


/***** file operations on device node *****/

static int martian_open (struct inode *inode, struct file *filp)
{
	// allow single open
	struct proc_dir_entry *entry = PDE (inode);
	struct martian *mdev = (struct martian *) entry->data;

	if (test_and_set_bit (MARTIAN_STATE_OPEN, &mdev->state)) {
		// already opened
		return -EBUSY;
	}

	filp->private_data = mdev;
	
	clear_bit (KMSTATE_FIFO_FAULT, &mdev->state);
	atomic_set (&mdev->count, 0);

	return 0;
}

/* clean up
 */
static int martian_release (struct inode *inode, struct file *filp)
{
	// struct martian *mdev = (struct martian *) PDE(inode)->data;
	struct martian *mdev = (struct martian *) filp->private_data;
	int debug;

#if defined (CONFIG_SMP) && (defined (__i386__) || defined (__x86_64__))
	if (test_and_clear_bit (KMSETTING_TRUESMP, &mdev->state)) {
		/* what if martian_modem screwed up but managed to survive */
		mspinlock_t *lock = &mdev->common->lock;
		if (unlikely (! mspin_trylock (lock))) {
			udelay (10);
			if (unlikely (! mspin_trylock (lock))) 
				MERROR ("Unfinished lock detected. Fixing\n");
		}
		mspin_unlock (lock);
	}
#endif

	atomic_set (&mdev->common->UnservedIrqs, 0);

	clear_bit (MARTIAN_STATE_OPEN, &mdev->state);
	clear_bit (KMSETTING_SERVE_IRQS, &mdev->state);
	debug = test_and_clear_bit (KMSETTING_DEBUG, &mdev->state);

	clear_bit (KMSTATE_UREADY, &mdev->state);
	MINFO ("martian_modem is detached\n");

	if (debug) {
		MFIFO_DUMP (mdev, dce_rx);
		MFIFO_DUMP (mdev, pdm_rx);
		MFIFO_DUMP (mdev, dce_tx);
	}

	atomic_set (&mdev->count, 0);
	/* check with wq */
	mars_write_register (mdev->common, IR, 0xff);

	return 0;
}

/* poll&select
 */
static unsigned int martian_poll(struct file *filp, struct poll_table_struct *wait)
{
	struct martian *mdev = (struct martian *) filp->private_data;

	if (atomic_read (& mdev->count) > 0)
		return POLLIN | POLLRDNORM; /* readable */

	poll_wait (filp, & mdev->wq, wait);

	if (atomic_read(& mdev->count) > 0)
		return POLLIN | POLLRDNORM; /* readable */

	return 0;
}

/* ioctl logger(s) */

#define DUMP_CORE_COMMON(fmt, var) MINFO (fmt, #var, common->var)
void dump_core_vars_common (struct martian_common *common) {
	MINFO ("Core vars:\n");
	MINFO ("Variable\t\t\tValue\n");

	DUMP_CORE_COMMON ("%s\t\t\t0x%x\n", BaseAddress);
	DUMP_CORE_COMMON ("%s\t\t\t0x%x\n", BaseAddress2);

	DUMP_CORE_COMMON ("%s\t\t0x%x\n", BaseAddressIndex);
	DUMP_CORE_COMMON ("%s\t\t\t0x%x\n", BaseValue);

	DUMP_CORE_COMMON ("%s\t\t\t0x%x\n", BaseAddressData);
	DUMP_CORE_COMMON ("%s\t\t\t0x%x\n", dp_bamil_rd7);
	DUMP_CORE_COMMON ("%s\t\t\t%d\n", x_dsp_mars);
}

/* ioctl getters */

static void fill_martian_conf (struct martian_conf *conf, struct martian *mdev) 
{
	struct martian_metrics *metrics = &mdev->metrics;

	conf->BaseAddress = metrics->BaseAddress;
	conf->CommAddress = metrics->CommAddress;
	conf->irq	  = metrics->irq;
	
	conf->device	  = metrics->device;
	conf->vendor	  = metrics->vendor;
	conf->subsys_vendor  = metrics->subsystem_vendor;
	conf->subsys_device  = metrics->subsystem_device;

	conf->flags	  = metrics->flags;

	conf->index	  = mdev->index;
}

static void 
fill_martian_irq (struct martian_irq_state *state, struct martian *mdev) 
{
	memcpy (state->string, irqstring, sizeof state->string);
	state->count  = interrupt_counter;	
	state->counts = mdev->icounts;
}

/*
 * sys_ioctl:
 * 	HAVE_UNLOCKED   -> unlocked_ioctl
 * 	! HAVE_UNLOCKED -> ioctl
 * 
 * compat_sys_ioctl
 * 	64, HAVE_COMPAT     -> compat_ioctl
 * 	64, ! HAVE_COMPAT   -> conversion; ioctl
 *
 * HAVE_UNLOCKED & HAVE_COMPAT -> new_ioctl for both unlocked & compat
 * !HAVE_UNLOCKED -> trad_ioctl for ioctl & compat
 *
 * dependency
 * (compat|unlocked)_ioctl: HAVE_UNLOCKED || HAVE_COMPAT
 * conventional_ioctl: !HAVE_UNLOCKED || !HAVE_COMPAT
 * assumed !def(HAVE_COMPAT) -> to be executed through conventional
 */

#if defined (HAVE_UNLOCKED_IOCTL) || defined (HAVE_COMPAT_IOCTL)
typedef long martian_ioctl_result_t;
#else
typedef int martian_ioctl_result_t;
#endif 	

#define copy_struct_to_user(dst, s) 		\
	copy_to_user ((void __user *) dst, &s, sizeof s) ? -EFAULT : 0

/*  martian_ioctl:
 *	primarily for unlocked and compat ioctl calls, 
 *	convention routed here too
 *
 *	no presumption about BKL
 */
static martian_ioctl_result_t 
martian_ioctl (struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct martian *mdev = (struct martian *) filp->private_data;
	/* void *buf; int size; */
	

	switch (cmd) {
	case MARTIAN_CONF_GET:
	{
		struct martian_conf conf;

		fill_martian_conf (&conf, mdev);
		return copy_struct_to_user (arg, conf);
	}
	case MARTIAN_IRQ_STATE:
	{
		struct martian_irq_state irq_state;

		fill_martian_irq (&irq_state, mdev);
		return copy_struct_to_user (arg, irq_state);
	}
	case MARTIAN_DUMP_VARS:
		dump_core_vars_common (mdev->common);
		return 0;

	case MARTIAN_SERVE_IRQS:
		if (! test_bit (KMSETTING_SERVE_IRQS, &mdev->state)) {
			MINFO ("serving irqs in module\n");

			mfifo_flush (&mdev->dce_rx);
			mfifo_flush (&mdev->dce_tx);
			mfifo_flush (&mdev->pdm_rx);

			set_bit (KMSETTING_SERVE_IRQS, &mdev->state);
		}
		return 0;

	case MARTIAN_MODEM_READY:
		set_bit (KMSTATE_UREADY, &mdev->state);
		MINFO ("martian_modem is attached.\n");
		return 0;


	case MARTIAN_KDEBUG:
		set_bit (KMSETTING_DEBUG, &mdev->state);
		MINFO ("debug option is enabled.\n");
		return 0;

	case MARTIAN_TRUE_SMP:
#ifdef CONFIG_SMP
#  if defined (__i386__) || defined (__x86_64__)
		if (test_bit (KMSETTING_TRUESMP, & mdev->state)) 
			return 0;

		ASSERT (! mspin_is_locked (& mdev->common->lock), 
				MERROR ("Assertion: dead lock\n"); 
				mspin_unlock (& mdev->common->lock);
		       );
		set_bit (KMSETTING_TRUESMP, &mdev->state);

		MINFO ("selected true smp.\n");
		return 0;
#  else
		MERROR ("could not run in true smp for this architecture\n");
		return -ENOIOCTLCMD;
#  endif

#else
		MERROR ("no true smp on UP kernel\n");
		return -ENOIOCTLCMD;
#endif

	case MARTIAN_CHECK_SMP:
#ifdef CONFIG_SMP
		MINFO ("cpu %ld is %sonline\n", arg, cpu_online (arg) ? "" : "not ");
		MINFO ("cpu_online_map|32 -- 0x%lx.\n", cpu_online_map.bits[0]);
		return 0;
#else
		MERROR ("cpu_online not available to modules on UP kernel\n");
#endif

	case MARTIAN_SCREWUP_DCE_RX:
		mdev->common->io_dce_rx_rptr = 0;
		MINFO ("setting dce_rx_rptr to 0\n");
		return 0;

	case MARTIAN_FIFO_FAULT:
		set_bit (KMSTATE_FIFO_FAULT, &mdev->state);
		MINFO ("imitating fifo fault\n");
		return 0;

	default:
		return -ENOIOCTLCMD;
	}
	/* return copy_to_user ((void __user *)arg, buf, size) ? -EFAULT : 0; */
}	



#if ! defined (HAVE_UNLOCKED_IOCTL) || ! defined (HAVE_COMPAT_IOCTL)

/* martian_conventional_ioctl:
 *	for older kernels
 *	routes to martian_ioctl
 */
static int 
martian_conventional_ioctl (struct inode *inop, struct file *filp, 
		unsigned int cmd, unsigned long arg) 
{
	/* we do shortening conversion here */
	return martian_ioctl (filp, cmd, arg);
}

#endif

#if   LINUX_VERSION_CODE  <  KERNEL_VERSION (2, 6, 10)
static inline int 
remap_pfn_range (struct vm_area_struct *vma, unsigned long addr,
                    unsigned long pfn, unsigned long size, pgprot_t prot)
{ 
	vma->vm_flags |= (VM_IO | VM_SHM | VM_LOCKED | VM_RESERVED);
	return remap_page_range (vma, addr, pfn << PAGE_SHIFT, size, prot);
}

#endif
 
/* martian_mmap: 
 *	maps common data into martian_modem process space
 *
 * 	this is very similar to drivers/char/mmtimer.c
 * 	we allow to write though
 *
 * key point:
 *	common page should be marked Reserved 
 *	otherwise we would finish with handpicked zero filled pages in martian_modem
 * 		function in question remap_pte_range 
 */
static int martian_mmap(struct file *filp, struct vm_area_struct * vma)
{
	struct martian *mdev = (struct martian *) filp->private_data;

	if (vma->vm_end - vma->vm_start != MCB_PAGES << PAGE_SHIFT)
		return -EINVAL;

	if (PAGE_SIZE > (1 << 16))
		return -ENOSYS;

#ifdef pgprot_noncached
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#endif
	if (remap_pfn_range(
		vma, 
		vma->vm_start, 
		__pa(mdev->common) >> PAGE_SHIFT,
		MCB_PAGES << PAGE_SHIFT, 
		vma->vm_page_prot
	    )) 
	{
		MERROR ("remap_pfn_range failed\n");
		return -EAGAIN;
	}

#if   LINUX_VERSION_CODE  <  KERNEL_VERSION (2, 6, 15)
	if (test_bit (KMSETTING_DEBUG, &mdev->state)) {
		vma->vm_flags &= ~VM_IO;
#ifdef VM_PFNMAP
		vma->vm_flags &= ~VM_PFNMAP;
#endif
	}
#endif
	return 0;
}

/* single reader only; 
 * it won't function correctly with more
 */
static ssize_t 
martian_read (struct file *filp, char __user *buf,
			size_t count, loff_t *ppos)
{
	struct martian *mdev = (struct martian *) filp->private_data;
	unsigned int pulses;
	int ret;

	if (count < sizeof mdev->count)
		return -EINVAL;
	
 	// NONBLOCK to be added

	ret = wait_event_interruptible (
		mdev->wq, 
 		(pulses = atomic_read (&mdev->count)) > 0 || test_bit (KMSTATE_FIFO_FAULT, &mdev->state)
		);

	if (ret == 0) {
		atomic_dec (&mdev->common->UnpassedIrqs);
		atomic_sub (pulses, &mdev->count); 
 		if (test_bit (KMSTATE_FIFO_FAULT, &mdev->state)) {
 			MDEBUG ("read: pulses 0x%x\n", pulses);
 			MINFO ("read: FIFO_FAULT was detected\n");
 			pulses |= 0x1000;
 			clear_bit (KMSTATE_FIFO_FAULT, &mdev->state);
 		}
 		if (put_user (pulses, (unsigned int __user *) buf) != 0)
			return -EFAULT;

		*ppos += sizeof pulses;
		return sizeof pulses;
	}

	return ret;
}

struct file_operations martian_fops = {
	.owner		= THIS_MODULE,
	.open		= martian_open,
	.release	= martian_release,
	
	.read		= martian_read,
	.poll		= martian_poll,

#if   ! defined (HAVE_UNLOCKED_IOCTL) || ! defined (HAVE_COMPAT_IOCTL)
	.ioctl		= martian_conventional_ioctl,
#endif

#if   HAVE_UNLOCKED_IOCTL
	.unlocked_ioctl	= martian_ioctl,
#endif

#if   HAVE_COMPAT_IOCTL
	.compat_ioctl	= martian_ioctl,
#endif

	.mmap		= martian_mmap
};

/***** martian devices management *****/

static int dev_idx = 0;

#ifndef PCMCIA
static void __devinit
#else
static void
#endif 
martian_configure_isr (struct martian *mdev) 
{
	struct martian_common *mcb = mdev->common;

	clear_bit (KMSTATE_UREADY, &mdev->state);
	clear_bit (KMSETTING_SERVE_IRQS, &mdev->state);
	init_waitqueue_head (& mdev->wq);
	atomic_set (&mdev->count, 0);

	/* register settings */
	mcb->BaseAddress	= mdev->metrics.BaseAddress;
	mcb->BaseAddressData	= mdev->metrics.BaseAddress + 1;
	mcb->BaseAddress2	= mdev->metrics.BaseAddress + 2;

	mcb->BaseAddressIndex	= mdev->metrics.BaseAddress;
	/* mcb->BaseValue - defined by first nonint reg{read,write} */

	mars_write_register (mcb, IR, 0xff);
}

struct martian *martian_add (struct martian_metrics *metrics, int *error)
{
	struct martian *mdev;
	int ret;
	char name[20] = "driver/mars/0";
	struct martian_common *mcb = NULL;

	MINFO("martian_add\n");
	// allocate device structure
	mdev = kmalloc (sizeof (struct martian), GFP_KERNEL);
	if (!mdev) {
		*error = -ENOMEM;
		return NULL; 
	}

	memset (mdev, 0, sizeof (struct martian));

	//mdev->metrics = *metrics;
	memcpy(&mdev->metrics, metrics, sizeof(struct martian_metrics));

	/*** allocate pages for common data ***/
	mcb = (void *) __get_free_pages (GFP_KERNEL, get_order (sizeof *mcb));
	if (! mcb) {
		MERROR ("page allocation failure\n");
		ret = -ENOMEM;
		goto err_free_mdev;
	}

	// as soon interrupt allocated we should be ready to handle it
	mdev->common = mcb;
	memset (mcb, 0, sizeof *mcb);
	martian_configure_isr (mdev);

	// claim resource
	ret = request_irq(
		metrics->irq, 
		martian_isr,  
		IRQF_SHARED,
		"164x", 
		mdev
	);

	if (ret < 0) {
		MERROR ("failed to request irq\n");
		goto err_free_common;
	}

	// just to be explicit
	clear_bit (MARTIAN_STATE_OPEN, &mdev->state);
	clear_bit (KMSETTING_DEBUG, &mdev->state);
	
	// create proc entry for this device
	name[ strlen(name) - 1 ] = '0' + dev_idx;

	mdev->entry = create_proc_entry (name, 0400, NULL);
	if (! mdev->entry) {
		MERROR ("martian: failed to create /proc/%s\n", name);
		ret = -ENOMEM;
		goto err_free_irq;
	}

	{
		int page, pages = (PAGE_SIZE - 1 + sizeof *mcb) >> PAGE_SHIFT;
		for (page = 0; page < pages; page++)
			SetPageReserved (
			  virt_to_page ((page << PAGE_SHIFT) + (char *) mcb)
			);
	}

	strcpy (mcb->signature, MARTIAN_SIGNATURE);
	strcpy (mcb->tsignature, MARTIAN_TSIGNATURE);
	strcpy (mcb->kstamp, __stringify (KMARTIAN_STAMP));

#if defined (CONFIG_SMP) && (defined (__i386__) || defined (__x86_64__))
	mspin_lock_init (&mcb->lock);
#endif

	MFIFO_INIT (&mdev->dce_rx, mcb->io_dce_rx_buff);
	MFIFO_INIT (&mdev->dce_tx, mcb->io_dce_tx_buff);
	MFIFO_INIT (&mdev->pdm_rx, mcb->io_pdm_rx_buff);

	mcb->tx_incorrespondence = 
	mcb->rx_incorrespondence =
	mcb->pdm_rx_incorrespondence = 0;

	mcb->pdm_rx_overrun	=
	mcb->rx_overrun		= 0;

	// set an index
	mdev->index = dev_idx;

	mdev->entry->proc_fops = & martian_fops;
	mdev->entry->data 	 = mdev;

	MINFO ("added device %04x:%04x " 
			"BaseAddress = 0x%x, CommAddres = 0x%x, irq = %d\n", 
			metrics->vendor, metrics->device, 
			metrics->BaseAddress, metrics->CommAddress,
			metrics->irq
		);
	printk(KERN_INFO "common block: %s, %s, %s\n",
		  mcb->signature,
		  mcb->tsignature,
		  mcb->kstamp);

	dev_idx++;

	return mdev;

err_free_irq:
	free_irq (metrics->irq, mdev);

err_free_common:
	free_pages ((unsigned long ) mcb, get_order (sizeof *mcb));

err_free_mdev:
	kfree (mdev);

	*error = ret;
	return NULL;
}


void martian_del (struct martian *mdev)
{
	{
		char name[20] = "driver/mars/0";
		name[ strlen(name) - 1 ] = '0' + mdev->index;

		remove_proc_entry (name, NULL);
	}
	free_irq (mdev->metrics.irq, mdev);
	{
		struct martian_common *mcb = mdev->common;
		int page, pages = (PAGE_SIZE - 1 + sizeof *mcb) >> PAGE_SHIFT;
		for (page = 0; page < pages; page++)
			ClearPageReserved (
			  virt_to_page ((page << PAGE_SHIFT) + (char *) mcb)
			);
		
		free_pages ((unsigned long ) mcb, get_order (sizeof *mcb));
	}

	kfree (mdev);
}

#ifndef PCMCIA
/***** PCI Driver Interface *****/

static int __devinit 
martian_device_gatherinfo (const struct pci_dev *dev, const struct pci_device_id *ent, struct martian_metrics *metrics)
{
	int bar;

	/*** common setting ***/
	metrics->subsystem_device = dev->subsystem_device;
	metrics->subsystem_vendor = dev->subsystem_vendor;

	metrics->irq = dev->irq;

	if (ent && ent->subvendor == PCI_VENDOR_ID_ATT) {
		/* Toshibas */

		switch (dev->class >> 8) {
		case PCI_CLASS_COMMUNICATION_SERIAL:
		case PCI_CLASS_COMMUNICATION_MULTISERIAL:
		case PCI_CLASS_COMMUNICATION_MODEM:
		case PCI_CLASS_COMMUNICATION_OTHER:
			break;

		default:
			return -ENODEV;
		}

		if (PCI_FUNC (dev->devfn) != 0) {
			MERROR ("Toshiba outside the range\n");
			return -ENODEV;
		}

		metrics->device = 0x440;
		metrics->vendor = PCI_VENDOR_ID_ATT;

		metrics->CommAddress = pci_resource_start (dev, 1);
		metrics->BaseAddress = pci_resource_start (dev, 2);
		metrics->flags |= MARTIAN_DEV_TOSHIBA;
		MDEBUG ("Toshiba device\n");

		return 0;
	}

	// check range for xircom
	if (ent && ent->vendor == PCI_VENDOR_ID_XIRCOM 
				&& ent->device == PCI_ANY_ID)

		switch (dev->class >> 8) {
		case PCI_CLASS_COMMUNICATION_SERIAL:
		case PCI_CLASS_COMMUNICATION_MULTISERIAL:
		case PCI_CLASS_COMMUNICATION_MODEM:
		case PCI_CLASS_COMMUNICATION_OTHER:
			if (! (dev->device <= 0x3ff)) {
				MERROR ("XIRCOM outside the range\n");
				return -ENODEV;
			}
			break;

		default:
			return -ENODEV;
		}
	
	// all ids added through sysfs come here

	/* Base/Common Addresses */
	for (bar = 5; bar >= 0; bar--) 
		if ( pci_resource_start (dev, bar) != 0 && 
			(pci_resource_flags (dev, bar) & IORESOURCE_IO) ) {
			metrics->BaseAddress =  pci_resource_start (dev, bar);
			break;
		}

	// continue
	for (bar--; bar >= 0; bar--) 
		if ( pci_resource_start (dev, bar) != 0 && 
			(pci_resource_flags (dev, bar) & IORESOURCE_IO) ) {
			metrics->CommAddress =  pci_resource_start (dev, bar);
			break;
		}

	if (! metrics->BaseAddress) {
		MERROR ("base address not found\n");
		return -ENODEV;
	}

	metrics->device = dev->device;
	metrics->vendor = dev->vendor;

	return 0;
}

static int __devinit 
mars_device_probe (struct pci_dev *dev, const struct pci_device_id *ent)
{
	int ret;
	struct martian *mdev;
	struct martian_metrics metrics;

	ret = pci_enable_device (dev);
	if (ret != 0) {
		MERROR ("failed to enable device\n");
		MDEBUG ("res %d\n", ret);
		return -ENODEV;
	}

	// initialize data
	ret = martian_device_gatherinfo (dev, ent, &metrics);
	if (ret < 0) {
		MERROR ("failed to gather device info\n");
		goto err_disable_pci;
	}

	mdev = martian_add (&metrics, &ret);
	if (! mdev) {
		goto err_disable_pci;
	}

	pci_set_drvdata (dev, mdev);

	dev_idx++;

	return 0;

err_disable_pci:
	pci_disable_device (dev);

	return ret;
}


static void __devexit mars_remove (struct pci_dev *dev) 
{
	martian_del (pci_get_drvdata (dev));
	pci_disable_device (dev);
}

#include "martian_ids.c"

// the device
struct pci_driver martian_driver = {
	.name		= "martian",
	.probe		= mars_device_probe,
	.remove		= __devexit_p (mars_remove),
	.id_table	= martian_ids
};

#define for_each_pci_dev_in_class(_class, mask, dev)  \
while ((dev = pci_get_device (PCI_ANY_ID, PCI_ANY_ID, dev)) != NULL) \
	if ((dev->class ^ _class) & mask)  	\
		continue;			\
	else

// ^ pci_get_class won't do cause we need a mask 

#if   LINUX_VERSION_CODE  <  KERNEL_VERSION (2, 6, 13)
#  define pci_match_id(ids, dev)	pci_match_device (ids, dev)
#endif

#define SERIAL_DRIVER	"serial"

static void detach_from_serial_class (unsigned int class, unsigned int mask) 
{
	struct pci_dev *dev = NULL;
	const struct pci_device_id *entry;


	for_each_pci_dev_in_class (class, mask, dev) {
		/*
		MDEBUG ("martian: detach: checking %x:%x by %s\n", 
			dev->vendor, dev->device, 
			dev->driver ? 
			  (dev->driver->name ? dev->driver->name : "unknown") :
			  "none"
		);
		*/

		if (! dev->driver || ! dev->driver->name || strcmp (SERIAL_DRIVER, dev->driver->name) ) 
			continue;

		entry = pci_match_id (martian_ids, dev);
		if (entry && entry->vendor != PCI_ANY_ID) {
			MDEBUG ("detaching %x:%x from %s\n", 
				dev->vendor, dev->device, dev->driver->name);

			device_release_driver (& dev->dev);
		}
	}
}

static void detach_from_serial(void) 
{
	detach_from_serial_class (PCI_CLASS_COMMUNICATION_SERIAL << 8, 0xffff00);
	detach_from_serial_class (PCI_CLASS_COMMUNICATION_MODEM << 8, 0xffff00);
	detach_from_serial_class (PCI_CLASS_COMMUNICATION_MULTISERIAL << 8, 0xffff00);
	detach_from_serial_class (PCI_CLASS_COMMUNICATION_OTHER << 8, 0xffff00);
}
#endif /*PCMCIA*/

/***** kernel module interface *****/
#ifndef PCMCIA
static int __init martian_init (void) 
#else
int martian_init(void)
#endif
{
	printk (KERN_INFO "martian loaded - " __stringify (KMARTIAN_STAMP) "\n");

	// allocate proc directory for martians
	martians_proc_dir = proc_mkdir ("driver/mars", NULL);
	if (! martians_proc_dir) {
		MERROR ("cannot create dir\n");
		return -ENOMEM;
	}

	// x86_64 ioctl32 should be conversed
#if defined (CONFIG_COMPAT) &&  ! defined (HAVE_COMPAT_IOCTL)
	register_ioctl32_conversion (MARTIAN_CONF_GET,  NULL);
	register_ioctl32_conversion (MARTIAN_IRQ_STATE, NULL);
	register_ioctl32_conversion (MARTIAN_DUMP_VARS, NULL);
#endif

#ifndef PCMCIA
	detach_from_serial();

#if   LINUX_VERSION_CODE  >=  KERNEL_VERSION (2, 6, 10)
	return pci_register_driver (&martian_driver);
#else
	return pci_module_init (&martian_driver);
#endif
#else
	return 0;
#endif /*PCMCIA*/
}

#ifndef PCMCIA
static void __exit martian_exit(void) 
#else
void martian_exit(void)
#endif
{
#ifndef PCMCIA
	pci_unregister_driver (&martian_driver);
#endif	
	remove_proc_entry ("driver/mars", NULL);
}

#ifndef PCMCIA
module_exit (martian_exit);

module_init (martian_init);

/***** module info *****/

MODULE_LICENSE ("GPL");
MODULE_AUTHOR ("A. Chentsov");
MODULE_DESCRIPTION ("Agere 164x dsp driver");
MODULE_DEVICE_TABLE( pci, martian_ids );
#endif /*PCMCIA*/
