/*======================================================================

    A driver for PCMCIA Lucent modem devices

    ltmodem_cs.c
    based on serial_cs.c 1.134 2002/05/04 05:48:53
	and martian.c 	2005 by A. Chentsov

    The contents of this file are subject to the Mozilla Public
    License Version 1.1 (the "License"); you may not use this file
    except in compliance with the License. You may obtain a copy of
    the License at http://www.mozilla.org/MPL/

    Software distributed under the License is distributed on an "AS
    IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
    implied. See the License for the specific language governing
    rights and limitations under the License.

    The initial developer of the original code is David A. Hinds
    <dahinds@users.sourceforge.net>.  Portions created by David A. Hinds
    are Copyright (C) 1999 David A. Hinds.  All Rights Reserved.

    Alternatively, the contents of this file may be used under the
    terms of the GNU General Public License version 2 (the "GPL"), in which
    case the provisions of the GPL are applicable instead of the
    above.  If you wish to allow the use of your version of this file
    only under the terms of the GPL and not to allow others to use
    your version of this file under the MPL, indicate your decision
    by deleting the provisions above and replace them with the notice
    and other provisions required by the GPL.  If you do not delete
    the provisions above, a recipient may use your version of this
    file under either the MPL or the GPL.
    
======================================================================*/

#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
#include <linux/autoconf.h>
#else
#include <linux/config.h>
#endif

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/serial_core.h>
#include <linux/major.h>
#include <linux/pci.h>
#include <linux/device.h>
#include <asm/io.h>
#include <asm/system.h>

#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ciscode.h>
#include <pcmcia/ds.h>
#include <pcmcia/ss.h>
#include <pcmcia/cisreg.h>

#include "kmartian.h"

#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
module_param(pc_debug, int, 0644);
#define DEBUG(n, args...) if (pc_debug>(n)) printk(KERN_DEBUG args)
#else
#define DEBUG(n, args...)
#endif
static char *version = "martian_drv_cs.c 0.4 2009/01/02 (Joern Wuestenfeld)";

/*====================================================================*/

/* Parameters that can be set with 'insmod' */

/* Enable the speaker? */
static int do_sound = 1;

module_param(do_sound, int, 0444);

/*====================================================================*/

/* Table of multi-port card ID's */

struct multi_id{
    u_short	manfid;
    u_short	prodid;
    int		multi;		/* 1 = multifunction, > 1 = # ports */
};

/* Yet no multifunction cards supported / known */
static struct multi_id multi_id[] = {
{0,0,0}
};
#define MULTI_COUNT (sizeof(multi_id)/sizeof(multi_id));

struct ltmodem_info{
    struct pcmcia_device	*p_dev;
    int		ndev;
    int		multi;
    int		slave;
    int		manfid;
    int		prodid;
    dev_node_t	node[4];
    int		line[4];
	struct martian *mdev;
};

static int martian_config(struct pcmcia_device *link);

static int martian_probe(struct pcmcia_device *link);
static void martian_detach(struct pcmcia_device *link);

/*static struct pcmcia_device *dev_list = NULL;*/

/* Functions for Lucent modem in martian.o */
extern int martian_init(void);
extern void martian_exit(void);
extern struct martian *martian_add (struct martian_metrics *metrics, int *error);
extern void martian_del(struct martian *mdev);
extern void mars_remove(struct pci_dev *dev);

/*====================================================================

    After a card is removed, martian_remove() will unregister
    the serial device(s), and release the PCMCIA configuration.

// // ======================================================================*/

static void martian_remove(struct pcmcia_device *link)
{
	struct ltmodem_info *info = link->priv;

	DEBUG(0, "martian_remove(0x%p)\n", link);

	/*
	 * Recheck to see if the device is still configured.
	 */
	martian_del(info->mdev);
		
	martian_exit();

	info->p_dev->dev_node = NULL;

	if (!info->slave)
		pcmcia_disable_device(link);
}

static int martian_suspend(struct pcmcia_device *link)
{
// 	link->state |= DEV_SUSPEND;
// 
// 	if (link->state & DEV_CONFIG) {
// 		struct ltmodem_info *info = link->priv;
// 		int i;
// 
// 		for (i = 0; i < info->ndev; i++)
// 			serial8250_suspend_port(info->line[i]);
// 
// 		if (!info->slave)
// 			pcmcia_release_configuration(link->handle);
// 	}
	return 0;
}

static int martian_resume(struct pcmcia_device *link)
{
// 	link->state &= ~DEV_SUSPEND;
// 
// 	if (DEV_OK(link)) {
// 		struct ltmodem_info *info = link->priv;
// 		int i;
// 
// 		if (!info->slave)
// 			pcmcia_request_configuration(link->handle, &link->conf);
// 
// 		for (i = 0; i < info->ndev; i++)
// 			serial8250_resume_port(info->line[i]);
// 	}
	return 0;
}

/*======================================================================

    martian_probe() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

======================================================================*/

static int martian_probe(struct pcmcia_device *link)
{
	struct ltmodem_info *info;

	DEBUG(0, "martian_attach()\n");

	/* Create new serial device */
	info = kzalloc(sizeof(*info), GFP_KERNEL);
		if (!info) return -ENOMEM;
	memset(info, 0, sizeof(*info));
	info->p_dev = link;
	link->priv = info;

	link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
	link->io.NumPorts1 = 8;
	link->irq.Attributes = IRQ_TYPE_DYNAMIC_SHARING;
	link->irq.IRQInfo1 = IRQ_SHARE_ID;
	link->irq.Handler = NULL;
	link->conf.Attributes = 0;
	if (do_sound) {
		link->conf.Attributes |= CONF_ENABLE_SPKR;
		link->conf.Status = CCSR_AUDIO_ENA;
	}
	link->conf.IntType = INT_MEMORY_AND_IO;

	/* Register with Card Services */
	
	return martian_config(link);
} /* martian_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/

static void martian_detach(struct pcmcia_device *link)
{
    struct ltmodem_info *info = link->priv;

    DEBUG(0, "martian_detach(0x%p)\n", link);
	
	/*
	 * Ensure any outstanding scheduled tasks are completed.
	 */
	flush_scheduled_work();

	martian_remove(link);
	pcmcia_disable_device(link);

    kfree(info);

} /* martian_detach */

/*====================================================================*/

static int setup_ltmodem(struct ltmodem_info *info, unsigned int port, unsigned int com, int irq)
{
    //struct serial_struct ltmodem;
    int line;
	int res;
	struct martian_metrics metrics;
	struct martian *mdev;
	
	res = martian_init();
    if(res < 0)
    {
	   printk(KERN_NOTICE "Failed to initialize modem\n");
	   return res;
    }
	
	metrics.BaseAddress = port;
	metrics.CommAddress = com;
	metrics.irq = irq;
	metrics.vendor = info->manfid;
	metrics.device = info->prodid;
	metrics.subsystem_vendor = info->manfid ;
	metrics.subsystem_device = info->prodid;
	metrics.flags = 0x0;
	
	printk(KERN_INFO "vendor: %04x, device %04x\n",info->manfid,info->prodid);
	mdev = martian_add(&metrics, &res);
	if(!mdev)
		return ENODEV;
	info->mdev = mdev;

    line = 0;
    info->line[info->ndev] = line;
    sprintf(info->node[info->ndev].dev_name, "ttySM%d", line);
    info->node[info->ndev].major = 62;
    info->node[info->ndev].minor = 0x64+line;
    if (info->ndev > 0)
	info->node[info->ndev-1].next = &info->node[info->ndev];
    info->ndev++;
    return 0;
}

/*====================================================================*/

static int
first_tuple(struct pcmcia_device *handle, tuple_t * tuple, cisparse_t * parse)
{
	int i;
	i = pcmcia_get_first_tuple(handle, tuple);
	if (i != CS_SUCCESS)
		return CS_NO_MORE_ITEMS;
	i = pcmcia_get_tuple_data(handle, tuple);
	if (i != CS_SUCCESS)
		return i;
	return pcmcia_parse_tuple(handle, tuple, parse);
}

static int
next_tuple(struct pcmcia_device *handle, tuple_t * tuple, cisparse_t * parse)
{
	int i;
	i = pcmcia_get_next_tuple(handle, tuple);
	if (i != CS_SUCCESS)
		return CS_NO_MORE_ITEMS;
	i = pcmcia_get_tuple_data(handle, tuple);
	if (i != CS_SUCCESS)
		return i;
	return pcmcia_parse_tuple(handle, tuple, parse);
}

/*====================================================================*/

static int simple_config(struct pcmcia_device *link)
{
    static unsigned int base[5] = { 0x3f8, 0x2f8, 0x3e8, 0x2e8, 0x0 };
    struct ltmodem_info *info;
    tuple_t tuple;
    u_char buf[128];
    cisparse_t parse;
    cistpl_cftable_entry_t *cf = &parse.cftable_entry;
    config_info_t config;
    int i, j, try;

    DEBUG(0, "simple_config\n");
    info = link->priv;
    /* If the card is already configured, look up the port and irq */
    i = pcmcia_get_configuration_info(link, &config);
    if ((i == CS_SUCCESS) && (config.Attributes & CONF_VALID_CLIENT))
    {
	   unsigned int port = 0;
	   if ((config.BasePort2 != 0) && (config.NumPorts2 == 8))
       {
	     port = config.BasePort2;
	     info->slave = 1;
       } 
	   if (info->slave)
	     return setup_ltmodem(info, port, link->conf.ConfigBase, config.AssignedIRQ);
    }
    //link->conf.Vcc = config.Vcc;

    link->io.NumPorts1 = 8;
    link->io.NumPorts2 = 0;

    /* First pass: look for a config entry that looks normal. */
    tuple.TupleData = (cisdata_t *)buf;
    tuple.TupleOffset = 0;
    tuple.TupleDataMax = sizeof(buf) * sizeof(u_char);
    tuple.Attributes = 0;
    tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
    /* Two tries: without IO aliases, then with aliases */
    for (try = 0; try < 2; try++)
    {
      i = first_tuple(link, &tuple, &parse);
      while (i != CS_NO_MORE_ITEMS)
      {
	    if (i != CS_SUCCESS) goto next_entry;
	    if (cf->vpp1.present & (1<<CISTPL_POWER_VNOM))
		  link->conf.Vpp =  cf->vpp1.param[CISTPL_POWER_VNOM]/10000;
	    if ((cf->io.nwin > 0) && (cf->io.win[0].len == 8) && (cf->io.win[0].base != 0))
        {
		  link->conf.ConfigIndex = cf->index;
		  link->io.BasePort1 = cf->io.win[0].base;
		  link->io.IOAddrLines = (try == 0) ? 16 : cf->io.flags & CISTPL_IO_LINES_MASK;
		  i = pcmcia_request_io(link, &link->io);
		  if (i == CS_SUCCESS) goto found_port;
	    }

        next_entry:
	       i = next_tuple(link, &tuple, &parse);
      }
    }

    /* Second pass: try to find an entry that isn't picky about
       its base address, then try to grab any standard serial port
       address, and finally try to get any free port. */
    i = first_tuple(link, &tuple, &parse);
    while (i != CS_NO_MORE_ITEMS) {
	if ((i == CS_SUCCESS) && (cf->io.nwin > 0) &&
	    ((cf->io.flags & CISTPL_IO_LINES_MASK) <= 3)) {
	    link->conf.ConfigIndex = cf->index;
	    for (j = 0; j < 5; j++) {
		link->io.BasePort1 = base[j];
		link->io.IOAddrLines = base[j] ? 16 : 3;
		i = pcmcia_request_io(link, &link->io);
		if (i == CS_SUCCESS) goto found_port;
	    }
	}
	i = next_tuple(link, &tuple, &parse);
    }

found_port:
    if (i != CS_SUCCESS) {
		cs_error(link, RequestIO, i);
		return -1;
    }
	
	link->conf.ConfigIndex = cf->index;
 	link->conf.Attributes = CONF_ENABLE_IRQ;
// 	link->conf.Present = PRESENT_OPTION;
// 	
 	link->irq.Attributes = IRQ_TYPE_DYNAMIC_SHARING|IRQ_FIRST_SHARED;
//  	link->irq.Handler     = martian_isr;
//  	link->irq.Instance    = info->mdev;

    i = pcmcia_request_irq(link, &link->irq);
    if (i != CS_SUCCESS) {
		cs_error(link, RequestIRQ, i);
		link->irq.AssignedIRQ = 0;
    }
   	i = pcmcia_request_configuration(link, &link->conf);
    if (i != CS_SUCCESS) {
		cs_error(link, RequestConfiguration, i);
		return -1;
    }
	try = setup_ltmodem(info, link->io.BasePort1, link->conf.ConfigBase,link->irq.AssignedIRQ);
    /* Finally, report what we've done */
    printk(KERN_INFO "%s\n",version);
    printk(KERN_INFO "%s: index 0x%02x: ",/*Vcc %d.%d",*/
	   info->node[0].dev_name, link->conf.ConfigIndex); /*,
	   link->conf.Vcc/10, link->conf.Vcc%10);*/
    if (link->conf.Vpp)
	printk(", Vpp %d.%d", link->conf.Vpp/10, link->conf.Vpp%10);
    if (link->conf.Attributes & CONF_ENABLE_IRQ)
	printk(", irq %d", link->irq.AssignedIRQ);
    if (link->io.NumPorts1)
	printk(", io 0x%04x-0x%04x", link->io.BasePort1,
	       link->io.BasePort1+link->io.NumPorts1-1);
    if (link->io.NumPorts2)
	printk(" & 0x%04x-0x%04x", link->io.BasePort2,
	       link->io.BasePort2+link->io.NumPorts2-1);
    if (link->win)
	printk(", mem 0x%06lx-0x%06lx", (long unsigned int)cf->io.win[0].base,
	       (long unsigned int)(cf->io.win[0].base+cf->io.win[0].len-1));
	if(link->conf.ConfigBase)
		printk(", config port 0x%04x", link->conf.ConfigBase);
    printk("\n");
	return try;
}

static int multi_config(struct pcmcia_device *link)
{
    struct ltmodem_info *info = link->priv;
    tuple_t tuple;
    u_char buf[255];
    cisparse_t parse;
    cistpl_cftable_entry_t *cf = &parse.cftable_entry;
    config_info_t config;
    int i, base2 = 0;

    pcmcia_get_configuration_info(link, &config);
    //link->conf.Vcc = config.Vcc;

    tuple.TupleData = (cisdata_t *)buf;
    tuple.TupleOffset = 0;
	tuple.TupleDataMax = sizeof(buf) * sizeof(u_char);
    tuple.Attributes = 0;
    tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;

    /* First, look for a generic full-sized window */
    link->io.NumPorts1 = info->multi * 8;
    i = first_tuple(link, &tuple, &parse);
    while (i != CS_NO_MORE_ITEMS) {
	/* The quad port cards have bad CIS's, so just look for a
	   window larger than 8 ports and assume it will be right */
	if ((i == CS_SUCCESS) && (cf->io.nwin == 1) &&
	    (cf->io.win[0].len > 8)) {
	    link->conf.ConfigIndex = cf->index;
	    link->io.BasePort1 = cf->io.win[0].base;
	    link->io.IOAddrLines = cf->io.flags & CISTPL_IO_LINES_MASK;
	    i = pcmcia_request_io(link, &link->io);
	    base2 = link->io.BasePort1 + 8;
	    if (i == CS_SUCCESS) break;
	}
	i = next_tuple(link, &tuple, &parse);
    }

    /* If that didn't work, look for two windows */
    if (i != CS_SUCCESS) {
	link->io.NumPorts1 = link->io.NumPorts2 = 8;
	info->multi = 2;
	i = first_tuple(link, &tuple, &parse);
	while (i != CS_NO_MORE_ITEMS) {
	    if ((i == CS_SUCCESS) && (cf->io.nwin == 2)) {
		link->conf.ConfigIndex = cf->index;
		link->io.BasePort1 = cf->io.win[0].base;
		link->io.BasePort2 = cf->io.win[1].base;
		link->io.IOAddrLines = cf->io.flags & CISTPL_IO_LINES_MASK;
		i = pcmcia_request_io(link, &link->io);
		base2 = link->io.BasePort2;
		if (i == CS_SUCCESS) break;
	    }
	    i = next_tuple(link, &tuple, &parse);
	}
    }

    if (i != CS_SUCCESS) {
	cs_error(link, RequestIO, i);
	return -1;
    }

    i = pcmcia_request_irq(link, &link->irq);
    if (i != CS_SUCCESS) {
	cs_error(link, RequestIRQ, i);
	link->irq.AssignedIRQ = 0;
    }
    /* Socket Dual IO: this enables irq's for second port */
    if (info->multi && (info->manfid == MANFID_SOCKET)) {
	link->conf.Present |= PRESENT_EXT_STATUS;
	link->conf.ExtStatus = ESR_REQ_ATTN_ENA;
    }
    i = pcmcia_request_configuration(link, &link->conf);
    if (i != CS_SUCCESS) {
	cs_error(link, RequestConfiguration, i);
	return -1;
    }

    setup_ltmodem(info, link->io.BasePort1, link->conf.ConfigBase, link->irq.AssignedIRQ);
    /* The Nokia cards are not really multiport cards */
    if (info->manfid == MANFID_NOKIA)
	return 0;
    for (i = 0; i < info->multi-1; i++)
	setup_ltmodem(info, base2+(8*i), link->conf.ConfigBase, link->irq.AssignedIRQ);

    return 0;
}

/*======================================================================

    ltmodem_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    serial device available to the system.

======================================================================*/

static int martian_config(struct pcmcia_device *link)
{
    struct ltmodem_info *info = link->priv;
    tuple_t tuple;
    u_short buf[128];
    cisparse_t parse;
    cistpl_cftable_entry_t *cf = &parse.cftable_entry;
    int last_ret, last_fn;

    DEBUG(0, "martian_config(0x%p)\n", link);

    tuple.TupleData = (cisdata_t *)buf;
    tuple.TupleOffset = 0;
    tuple.TupleDataMax = 128;
    tuple.Attributes = 0;
    /* Get configuration register information */
    tuple.DesiredTuple = CISTPL_CONFIG;
    last_ret = first_tuple(link, &tuple, &parse);
    if (last_ret != CS_SUCCESS) {
	last_fn = ParseTuple;
	goto cs_failed;
    }
    link->conf.ConfigBase = parse.config.base;
    link->conf.Present = parse.config.rmask[0];

    /* Is this a compliant multifunction card? */
    tuple.DesiredTuple = CISTPL_LONGLINK_MFC;
    tuple.Attributes = TUPLE_RETURN_COMMON | TUPLE_RETURN_LINK;
    info->multi = (first_tuple(link, &tuple, &parse) == CS_SUCCESS);

    /* Another check for dual-serial cards: look for either serial or
       multifunction cards that ask for appropriate IO port ranges */
    tuple.DesiredTuple = CISTPL_FUNCID;
    if ((info->multi == 0) &&
	((first_tuple(link, &tuple, &parse) != CS_SUCCESS) ||
	 (parse.funcid.func == CISTPL_FUNCID_MULTI) ||
	 (parse.funcid.func == CISTPL_FUNCID_SERIAL))) {
	tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
	if (first_tuple(link, &tuple, &parse) == CS_SUCCESS) {
	    if ((cf->io.nwin == 1) && (cf->io.win[0].len % 8 == 0))
		info->multi = cf->io.win[0].len >> 3;
	    if ((cf->io.nwin == 2) && (cf->io.win[0].len == 8) &&
		(cf->io.win[1].len == 8))
		info->multi = 2;
	}
    }
    
    tuple.DesiredTuple = CISTPL_MANFID;
    if(first_tuple(link, &tuple, &parse) == CS_SUCCESS)
    {
		info->manfid = parse.manfid.manf;
		info->prodid = parse.manfid.card;
    }
    if (info->multi > 1)
	multi_config(link);
    else
	simple_config(link);

    if (info->ndev == 0)
	goto failed;

    link->dev_node = &info->node[0];
    return 0;

cs_failed:
    cs_error(link, last_fn, last_ret);
failed:
    martian_remove(link);
    return -ENODEV;

} /* martian_config */

/*====================================================================*/
static struct pcmcia_device_id martian_ids[] = {
	PCMCIA_DEVICE_MANF_CARD(0x018a,0x0001),
	PCMCIA_DEVICE_PROD_ID12("PCMCIA","V.90 Communication Device ",0x281F1C5D,0x0580AD93),
	PCMCIA_DEVICE_PROD_ID12("LT WIN MODEM", "PC card",0x10cede8e,0x5c88f9be),
	PCMCIA_DEVICE_NULL
};
MODULE_DEVICE_TABLE(pcmcia, martian_ids);

static struct pcmcia_driver martian_cs_driver = {
	.owner		= THIS_MODULE,
	.drv		= {
		.name	= "martian_drv_cs",
	},
	.probe		= martian_probe,
	.remove		= martian_detach,
	.id_table	= martian_ids,
	.suspend	= martian_suspend,
	.resume		= martian_resume,
};


static int __init init_martian_cs(void)
{
	return pcmcia_register_driver(&martian_cs_driver);
}

static void __exit exit_martian_cs(void)
{
    DEBUG(0, "martian_drv_cs: unloading\n");
    pcmcia_unregister_driver(&martian_cs_driver);
}

module_init(init_martian_cs);
module_exit(exit_martian_cs);

MODULE_AUTHOR("Joern Wuestenfeld");
MODULE_DESCRIPTION("Agere 164x dsp PCMCIA driver");
MODULE_LICENSE("GPL");


